<?php
/**
 * @author Vincent Petry <pvince81@owncloud.com>
 * @copyright 2017 ownCloud, Inc.
 *
 * This code is covered by the ownCloud Commercial License.
 *
 * You should have received a copy of the ownCloud Commercial License
 * along with this program. If not, see <https://owncloud.com/licenses/owncloud-commercial/>.
 *
 */

namespace OCA\Enterprise_key\Service;

use OCP\App\IAppManager;
use OCP\ILogger;
use OCP\IAppConfig;

class AppService {

	/**
	 * @var ILogger
	 */
	private $logger;

	/**
	 * @var IAppManager
	 */
	private $appManager;

	/**
	 * @var IAppConfig
	 */
	private $appConfig;

	public function __construct(
		ILogger $logger,
		IAppManager $appManager,
		IAppConfig $appConfig
	) {
		$this->logger = $logger;
		$this->appManager = $appManager;
		$this->appConfig = $appConfig;
	}

	/**
	 * Disable all enterprise apps
	 *
	 * @return string[] list of apps that got disabled
	 */
	public function disableEnterpriseApps() {
		$apps = $this->findEnterpriseApps();

		foreach ($apps as $app) {
			$this->logger->debug("Disabling enterprise app \"$app\"", ['app' => 'enterprise']);
			$this->appManager->disableApp($app);
		}

		return $apps;
	}

	/**
	 * Checks whether the give app is an enterprise app
	 *
	 * @param string $app app id
	 * @return bool true if it is an enterprise app, false otherwise
	 */
	private function isEnterpriseApp($app) {
		$info = $this->appManager->getAppInfo($app);
		$license = '';
		$author = '';

		if (isset($info['licence'])) {
			$license = $info['licence'];
		}
		if (isset($info['author'])) {
			$author = $info['author'];
		}
		
		if ($license === 'OCL' && preg_match('/owncloud /i', $author) === 1) {
			return true;
		}
	}

	/**
	 * Find all enabled enterprise apps
	 *
	 * @return string[] app id of enterprise apps
	 */
	private function findEnterpriseApps() {
		$apps = $this->appManager->getInstalledApps();

		// filter by enterprise apps only
		return array_filter($apps, function($app) {
			if ($app === 'enterprise_key') {
				// don't disable self
				return false;
			}

			// exclude already disabled apps
			if ($this->appConfig->getValue($app, 'enabled', 'no') === 'no') {
				return false;
			}

			if (!$this->isEnterpriseApp($app)) {
				return false;
			}

			return true;
		});
	}

}

